import io
from typing import Iterable, List, Union
from zipfile import ZipFile

from potyk_doc.models import File, FileData


def zip_files(files: Iterable[File], zip_name: str) -> File:
    """
    Создает zip-архив с файлами {files} и названием {zip_name}
    :param files: Список File
    :param zip_name: Название архива
    :return: File из байтов zip-архива и названия
    """
    stream = io.BytesIO()

    with ZipFile(stream, mode="w") as zf:
        for file_data, filename in files:
            zf.writestr(filename, file_data)

    return File(stream.getvalue(), zip_name)


def list_zip_files(zip_file: Union[File, FileData]) -> List[File]:
    """
    Выводит содержимое zip-архива
    :param zip_file: zip-архив в виде File или байтов
    :return: Содержимое zip-архива в виде списка File
    """

    def zip_file_generator(zf: ZipFile) -> Iterable[File]:
        for filename in zf.namelist():
            content = zf.read(filename)
            yield File(content, filename)

    if isinstance(zip_file, FileData):
        zip_data = zip_file
    else:
        zip_data, _ = zip_file

    zf = ZipFile(io.BytesIO(zip_data))

    return list(zip_file_generator(zf))



import datetime as dt
from dataclasses import dataclass
from typing import Union

from potyk_doc.translation.case import Case

MONTHS = {
    Case.Nominative: {
        1: "январь",
        2: "февраль",
        3: "март",
        4: "апрель",
        5: "май",
        6: "июнь",
        7: "июль",
        8: "август",
        9: "сентябрь",
        10: "октябрь",
        11: "ноябрь",
        12: "декабрь",
    },
    Case.Genitive: {
        1: "января",
        2: "февраля",
        3: "марта",
        4: "апреля",
        5: "мая",
        6: "июня",
        7: "июля",
        8: "августа",
        9: "сентября",
        10: "октября",
        11: "ноября",
        12: "декабря",
    },
    Case.Prepositional: {
        1: "январе",
        2: "феврале",
        3: "марте",
        4: "апреле",
        5: "мае",
        6: "июне",
        7: "июле",
        8: "августе",
        9: "сентябре",
        10: "октябре",
        11: "ноябре",
        12: "декабре",
    },
}


def date_to_ru(date_: dt.date, case: Case = Case.Genitive, without_day: bool = False) -> str:
    """
    >>> date_to_ru(dt.date(2019, 3, 11))
    '11 марта 2019 г.'
    >>> date_to_ru(dt.date(2019, 2, 14), Case.Prepositional, without_day=True)
    'феврале 2019 г.'
    """
    ru_date = RuDate(date_, case)
    if without_day:
        return ru_date.month_with_year

    return str(ru_date)


@dataclass()
class RuDate:
    """
    >>> ru_date = RuDate(dt.datetime(2019, 1, 22))
    >>> str(ru_date)
    '22 январь 2019 г.'
    >>> ru_date.month
    'январь'
    >>> ru_date.year
    '2019 г.'
    >>> ru_date.month_with_year
    'январь 2019 г.'
    >>> ru_date = RuDate(dt.datetime(2019, 1, 22), Case.Genitive)
    >>> str(ru_date)
    '22 января 2019 г.'
    """

    date_: Union[dt.date, dt.datetime]
    case: Case = Case.Nominative

    def __str__(self) -> str:
        return f"{self.date_.day} {self.month} {self.year}"

    @property
    def month(self) -> str:
        return MONTHS[self.case][self.date_.month]

    @property
    def year(self) -> str:
        return f"{self.date_.year} г."

    @property
    def month_with_year(self) -> str:
        """
        >>> ru_date = RuDate(dt.datetime(2019, 1, 22))
        >>> ru_date.month_with_year
        'январь 2019 г.'
        """
        return f"{self.month} {self.year}"



import numbers
import re
from dataclasses import dataclass
from decimal import Decimal
from typing import Union, Dict, Tuple, Any, cast, Optional
from potyk_lib.fp import If, pipe
from potyk_lib.num import decimal_round
from potyk_doc.money import Currency
from potyk_doc.translation.number import int_to_ru

CurrencyShorthands = {Currency.Rubles: "руб.", Currency.Tenge: "тг."}

RUBS: Dict[Tuple, str] = {(1,): "рубль", (2, 3, 4): "рубля", (0, 5, 6, 7, 8, 9): "рублей"}
PENNIES: Dict[Tuple, str] = {(1,): "копейка", (2, 3, 4): "копейки", (0, 5, 6, 7, 8, 9): "копеек"}


@dataclass()
class NumberAttrTranslator:
    """
    >>> from unittest.mock import MagicMock
    >>> summary = MagicMock(order_sum=3681.76, bool_=True)
    >>> NumberAttrTranslator(summary).order_sum
    '3 681 (Три тысячи шестьсот восемьдесят один) рубль 76 копеек'
    >>> NumberAttrTranslator(summary, attr_suffix='_str').bool_
    True
    >>> NumberAttrTranslator(summary, attr_suffix='_str').order_sum_str
    '3 681 (Три тысячи шестьсот восемьдесят один) рубль 76 копеек'
    >>> NumberAttrTranslator(summary, attr_suffix='_str').order_sum
    3681.76
    >>> NumberAttrTranslator(summary, attr_suffix='_str').bool_
    True

    """
    translatable: Any
    attr_suffix: Optional[str] = None

    def __getattr__(self, attr: str) -> str:
        is_number = lambda val: isinstance(val, numbers.Number) and not isinstance(val, bool)
        cut_suffix = lambda attr: re.sub(fr'{self.attr_suffix}$', '', attr)
        get_val = lambda attr: getattr(self.translatable, attr)
        translate = lambda val: sum_to_ru(val) if is_number(val) else val

        return (
            If(self.attr_suffix)
            .then(
                lambda: If(attr.endswith(self.attr_suffix))
                .then(lambda: pipe(cut_suffix, get_val, translate)(attr))
                .el(lambda: get_val(attr))
            )
            .el(lambda: pipe(get_val, translate)(attr))
            .get()
        )


def sum_to_ru(
    number: Union[int, float, Decimal],
    with_number: bool = True,
    currency: Currency = Currency.Rubles,
    with_pennies: bool = True,
) -> str:
    """
    >>> sum_to_ru(3681)
    '3 681 (Три тысячи шестьсот восемьдесят один) рубль 00 копеек'
    >>> sum_to_ru(13620)
    '13 620 (Тринадцать тысяч шестьсот двадцать) рублей 00 копеек'
    >>> sum_to_ru(3681, with_number=False)
    'Три тысячи шестьсот восемьдесят один рубль 00 копеек'
    >>> sum_to_ru(3681, with_pennies=False)
    '3 681 (Три тысячи шестьсот восемьдесят один) рубль'
    >>> sum_to_ru(1488, currency=Currency.Tenge)
    '1 488 (Одна тысяча четыреста восемьдесят восемь) тенге 00 тиын'
    >>> sum_to_ru(Decimal('-811.02'))
    '-811 (Минус восемьсот одиннадцать) рублей 02 копейки'
    """
    int_number = int(number)
    number_rest = cast(float, abs(number - int(number)))

    number_str = f"{int_number:,}".replace(",", " ")
    ru_number = int_to_ru(int_number)

    normalized_rest = normalize_pennies(number_rest)

    currency_str, pennies_str = _get_currency_names(currency, int_number, normalized_rest)

    if with_number:
        currency_str = f"{number_str} ({ru_number}) {currency_str}"
    else:
        currency_str = f"{ru_number} {currency_str}"

    if with_pennies:
        pennies_str = f" {normalized_rest} {pennies_str}"
    else:
        pennies_str = ""

    return f"{currency_str}{pennies_str}"


def _get_currency_names(
    currency: Currency, int_number: int, normalized_res: str
) -> Tuple[str, str]:
    if currency == Currency.Rubles:
        currency_str = create_ruble_str(int_number)
        pennies_str = create_pennies_str(normalized_res)
    elif currency == Currency.Tenge:
        currency_str = "тенге"
        pennies_str = "тиын"
    else:
        raise ValueError(f"Invalid currency: {currency}")

    return currency_str, pennies_str


def create_ruble_str(number: int) -> str:
    """
    >>> create_ruble_str(1)
    'рубль'
    >>> create_ruble_str(11)
    'рублей'
    >>> create_ruble_str(108)
    'рублей'
    >>> create_ruble_str(332)
    'рубля'
    >>> create_ruble_str(999)
    'рублей'
    """
    if number % 100 in range(11, 20):
        return "рублей"

    last_digit = number % 10
    return next(
        ruble_str for ruble_digits, ruble_str in RUBS.items() if last_digit in ruble_digits
    )


def normalize_pennies(number_rest: Union[float, Decimal]) -> str:
    """"
    >>> normalize_pennies(0)
    '00'
    >>> normalize_pennies(0.1)
    '10'
    >>> normalize_pennies(0.04)
    '04'
    >>> normalize_pennies(0.01)
    '01'
    >>> normalize_pennies(0.02)
    '02'
    """
    if not number_rest:
        rest_part = "00"
    else:
        rest_part = str(decimal_round(number_rest))[2:]
        if len(rest_part) == 1:
            rest_part += "0"

    return rest_part


def create_pennies_str(normalized_rest: str) -> str:
    """
    >>> create_pennies_str("00")
    'копеек'
    >>> create_pennies_str("10")
    'копеек'
    >>> create_pennies_str("04")
    'копейки'
    >>> create_pennies_str("01")
    'копейка'
    """
    if int(normalized_rest) % 100 in range(11, 20):
        return "копеек"

    return next(v for k, v in PENNIES.items() if int(normalized_rest[-1]) in k)



import inspect
from pathlib import Path
from typing import Union

from potyk_doc.models import FileName, FileData


def this_dir(caller_file: Union[str, Path] = None) -> Path:
    """
    Возвращает путь к директории модуля из которого запущена эта функция.

    Так, если запустить функцию в модуле tests/test_docx.py, то она вернет tests/.

    https://stackoverflow.com/a/8663885/5500609
    :return: Path на директорию модуля, в котором вызвана функция
    """
    caller_file = caller_file or _get_caller_file()
    return Path(caller_file).resolve().parent


def _get_caller_file(depth=2) -> str:
    """
    Возвращает путь к модулю, из которого был запущен этот метод
    """
    return inspect.stack()[depth][1]


def read_f(path, mode='rb'):
    with open(path, mode) as f:
        return f.read()


def read_f_in_this_dir(filename: FileName, mode='rb'):
    return read_f(this_dir(_get_caller_file()) / filename, mode)


def save_f_in_this_dir(filename: FileName, content: FileData):
    with open(this_dir(_get_caller_file()) / filename, mode='wb') as f:
        f.write(content)



import enum
import os.path
from typing import NamedTuple


# https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types
# https://en.wikipedia.org/wiki/Media_type
class Mimetype(str, enum.Enum):
    text = 'text/plain'
    html = 'text/html'
    csv = 'text/csv'
    json = 'application/json'
    pdf = 'application/pdf'
    zip = 'application/zip'
    any = 'application/octet-stream'
    docx = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
    xlsx = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'

    @classmethod
    def guess(cls, filename):
        """
        >>> Mimetype.guess('archive.zip')
        <Mimetype.zip: 'application/zip'>
        >>> Mimetype.guess('op.oppa.zip')
        <Mimetype.zip: 'application/zip'>
        """
        _, ext = os.path.basename(filename).rsplit('.', 1)
        return cls.__members__[ext]


class DocumentType(str, enum.Enum):
    pdf = 'pdf'
    docx = 'docx'

    @property
    def mime_type(self):
        """
        >>> DocumentType.pdf.mime_type
        'application/pdf'
        """
        return Mimetype.__members__[self.value].value


FileData = bytes
FileName = str


class File(NamedTuple):
    """Файл - тьюпл из байтов и названия файла"""
    data: FileData
    name: FileName


HTMLStr = str



"""
Ru translation of numbers
Source:
https://github.com/seriyps/ru_number_to_text
"""

from typing import List, Tuple

units = (
    u"ноль",
    (u"один", u"одна"),
    (u"два", u"две"),
    u"три",
    u"четыре",
    u"пять",
    u"шесть",
    u"семь",
    u"восемь",
    u"девять",
)

teens = (
    u"десять",
    u"одиннадцать",
    u"двенадцать",
    u"тринадцать",
    u"четырнадцать",
    u"пятнадцать",
    u"шестнадцать",
    u"семнадцать",
    u"восемнадцать",
    u"девятнадцать",
)

tens = (
    teens,
    u"двадцать",
    u"тридцать",
    u"сорок",
    u"пятьдесят",
    u"шестьдесят",
    u"семьдесят",
    u"восемьдесят",
    u"девяносто",
)

hundreds = (
    u"сто",
    u"двести",
    u"триста",
    u"четыреста",
    u"пятьсот",
    u"шестьсот",
    u"семьсот",
    u"восемьсот",
    u"девятьсот",
)

orders = (
    ((u"", u"", u""), "m"),
    ((u"тысяча", u"тысячи", u"тысяч"), "f"),
    ((u"миллион", u"миллиона", u"миллионов"), "m"),
    ((u"миллиард", u"миллиарда", u"миллиардов"), "m"),
)

minus = u"минус"


def num2text(num: int) -> str:
    """
    http://ru.wikipedia.org/wiki/Gettext#.D0.9C.D0.BD.D0.BE.D0.B6.D0.B5.D1.81.\
    D1.82.D0.B2.D0.B5.D0.BD.D0.BD.D1.8B.D0.B5_.D1.87.D0.B8.D1.81.D0.BB.D0.B0_2
    """
    if num == 0:
        return " ".join((units[0], orders[0][0][2])).strip()  # ноль

    rest = abs(num)
    ord = 0
    name = []
    while rest > 0:
        plural, nme = thousand(rest % 1000, orders[ord][1])
        if nme or ord == 0:
            name.append(orders[ord][0][plural])
        name += nme
        rest = int(rest / 1000)
        ord += 1
    if num < 0:
        name.append(minus)
    name.reverse()
    return " ".join(name).strip()


def thousand(rest: int, sex: str) -> Tuple[int, List[str]]:  # pylint: disable=too-complex
    """Converts numbers from 19 to 999"""
    prev = 0
    plural = 2
    name = []
    use_teens = 10 <= rest % 100 <= 19

    data: List[Tuple[Tuple, int]]
    if use_teens:
        data = [(teens, 10), (hundreds, 1000)]
    else:
        data = [(units, 10), (tens, 100), (hundreds, 1000)]

    for names, x in data:
        cur = int(((rest - prev) % x) * 10 / x)
        prev = rest % x
        if x == 10 and use_teens:
            plural = 2
            name.append(teens[cur])
        elif cur == 0:
            continue
        elif x == 10:
            name_ = names[cur]
            if isinstance(name_, tuple):
                name_ = name_[0 if sex == "m" else 1]
            name.append(name_)
            if 2 <= cur <= 4:
                plural = 1
            elif cur == 1:
                plural = 0
            else:
                plural = 2
        else:
            name.append(names[cur - 1])
    return plural, name



import dataclasses
import io
from pathlib import Path
from typing import Dict, Union

import PyPDF2
import pdfkit

from potyk_doc.models import HTMLStr, FileData


@dataclasses.dataclass()
class WkhtmltopdfOptions:
    """
    >>> WkhtmltopdfOptions().page_width('209.804').page_height("296.926").options
    {'--page-width': '209.804', '--page-height': '296.926'}
    """
    options: Dict[str, str] = dataclasses.field(default_factory=dict)

    def add_option(self, option_name, option_val):
        return dataclasses.replace(self, options={**self.options, option_name: option_val})

    def page_width(self, page_width_mm: str):
        return self.add_option('--page-width', page_width_mm)

    def page_height(self, page_height_mm: str):
        return self.add_option('--page-height', page_height_mm)

    def footer_html(self, footer_html_path: str):
        return self.add_option("--footer-html", footer_html_path)

    def header_html(self, header_html_path: str):
        return self.add_option("--header-html", header_html_path)

    def margin_bottom(self, margin_mm: str):
        margin_mm = margin_mm if margin_mm.endswith('mm') else f'{margin_mm}mm'
        return self.add_option("margin-bottom", margin_mm)


def render_pdf_from_html(
    pdf_html: HTMLStr,
    css_path: Union[str, Path, None] = None,
    options: Union[dict, WkhtmltopdfOptions, None] = None,
) -> FileData:
    """
    Рендерит pdf из html {pdf_html}.
    Рендер происходит с помощью либы pdfkit,
    которая в свою очередь использует `wkhtmltopdf <https://wkhtmltopdf.org/>`_ (=> она должна быть установлена)

    :param pdf_html: HTML-строка
    :param css_path: (опционально) путь к css-файлу, в котором будут стили, применяемые к html перед рендерингом
    :param options: (опционально) Словарь опций wkhtmltopdf, напр. {"--page-width": "209.804"}
    :return: pdf-байты
    """
    options = options.options if isinstance(options, WkhtmltopdfOptions) else options
    pdf_data = pdfkit.from_string(pdf_html, False, css=css_path, options=options)
    return pdf_data


def pdfs_are_equal(pdf_1: bytes, pdf_2: bytes) -> bool:
    pdf_1_data = PyPDF2.PdfReader(io.BytesIO(pdf_1))
    pdf_2_data = PyPDF2.PdfReader(io.BytesIO(pdf_2))
    return (
        len(pdf_1_data.pages) == len(pdf_2_data.pages) and
        all(
            pdf_1_page.extract_text() == pdf_2_page.extract_text()
            for pdf_1_page, pdf_2_page in zip(pdf_1_data.pages, pdf_2_data.pages)
        )
    )



import io
from pathlib import Path
from typing import Any, Union

import docx
import docx.document
from docxtpl import DocxTemplate


def render_docx_from_template(docx_path: Union[str, Path], **context: Any) -> bytes:
    """
    Рендерит docx из шаблона {docx_path}
    :param docx_path: Путь к docx-шаблону - docx-файл, использующий jinja-like синтаксис согласно `docxtpl <https://docxtpl.readthedocs.io/en/latest/>`_
    :param context: kwargs, которые передаются в шаблон (аналогично тому, как это делается в jinja)
    :return: docx-байты
    """
    doc = DocxTemplate(docx_path)
    doc.render(context)
    doc.save(stream := io.BytesIO())
    return stream.getvalue()


def docxs_are_equal(docx_1: bytes, docx_2: bytes) -> bool:
    def get_paragraph_data(doc: docx.document.Document) -> Any:
        return tuple(p.text for p in doc.paragraphs)

    def get_table_data(doc: docx.document.Document) -> Any:
        return tuple(
            tuple(cell.text for cell in row.cells)
            for table in doc.tables for row in table.rows
        )

    docx_1_data = docx.Document(io.BytesIO(docx_1))
    docx_2_data = docx.Document(io.BytesIO(docx_2))

    return (
        get_paragraph_data(docx_1_data) == get_paragraph_data(docx_2_data) and
        get_table_data(docx_1_data) == get_table_data(docx_2_data)
    )



import dataclasses
from typing import Dict
from urllib.parse import urlencode, quote

from potyk_doc.models import Mimetype


@dataclasses.dataclass
class ContentDisposition:
    filename: str
    mimetype: Mimetype = Mimetype.any

    @property
    def content_type(self):
        return self.mimetype.value

    @classmethod
    def docx(cls, filename):
        return cls(filename, Mimetype.docx)

    @classmethod
    def xlsx(cls, filename):
        return cls(filename, Mimetype.xlsx)

    @classmethod
    def pdf(cls, filename):
        return cls(filename, Mimetype.pdf)

    @classmethod
    def zip(cls, filename):
        return cls(filename, Mimetype.zip)

    @classmethod
    def guess(cls, filename):
        return cls(filename, Mimetype.guess(filename))

    @property
    def filename_encoded(self) -> str:
        """
        >>> ContentDisposition.xlsx('Отчет.xlsx').filename_encoded
        'filename=%D0%9E%D1%82%D1%87%D0%B5%D1%82.xlsx'
        """
        return urlencode({"filename": self.filename}, quote_via=quote)

    @property
    def header(self) -> str:
        """
        >>> ContentDisposition.xlsx('report.xlsx').header
        'attachment; filename=report.xlsx'
        """
        return f'attachment; {self.filename_encoded}'

    @property
    def header_dict(self) -> Dict[str, str]:
        """
        >>> ContentDisposition.xlsx('report.xlsx').header_dict
        {'Content-Disposition': 'attachment; filename=report.xlsx'}
        """
        return {'Content-Disposition': self.header}



from potyk_doc.translation._num2t4ru import num2text


def int_to_ru(number: int) -> str:
    """
    >>> int_to_ru(3681)
    'Три тысячи шестьсот восемьдесят один'
    """
    return num2text(number).capitalize()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['potyk_doc', 'potyk_doc.translation']

package_data = \
{'': ['*']}

install_requires = \
['PyPDF2>=3.0.0,<4.0.0',
 'docxtpl>=0.16.4,<0.17.0',
 'pdfkit>=1.0.0,<2.0.0',
 'potyk-lib>=0.5,<0.6']

setup_kwargs = {
    'name': 'potyk-doc',
    'version': '0.7.6',
    'description': '',
    'long_description': None,
    'author': 'potykion',
    'author_email': 'potykion@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
